/*-----------------------------------------------------------------------

File  : che_heuristic_parms.c

Author: Stephan Schulz

Contents
 
  Collection of parameters for heuristics and proofcontrol.


Changes

<1> Mon Jun  8 02:14:51 MET DST 1998
    New
<2> Wed Dec 16 18:47:49 MET 1998
    See .h file

-----------------------------------------------------------------------*/

#include "che_heuristic_parms.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/



/*-----------------------------------------------------------------------
//
// Function: HeuristicParmsAlloc()
//
//   Allocate a cell for parameters, with initialized empty stacks.
//
// Global Variables: -
//
// Side Effects    : Memory Operations
//
/----------------------------------------------------------------------*/

HeuristicParms_p HeuristicParmsAlloc()
{
   HeuristicParms_p handle = HeuristicParmsCellAlloc();
   
   handle->heuristic_name          = "Refinedweight";
   handle->wfcb_definitions        = PStackAlloc();
   handle->heuristic_definitions   = PStackAlloc();
   handle->ordertype               = NoOrdering;
   handle->to_weight_gen           = WNoMethod;
   handle->to_prec_gen             = PNoMethod;
   handle->to_const_weight         = WConstNoSpecialWeight;
   handle->use_indexing            = true;
   handle->forward_demod           = FullRewrite;
   handle->prefer_initial_clauses  = undefined;
   handle->paramod_with_units_only = undefined;   

   /* Obsolete old stuff, should vanish soon. */
   handle->i_parms                 = PStackAlloc();
   handle->d_parms                 = DStackAlloc();
   handle->s_parms                 = PStackAlloc();

   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: HeuristicParmsFree()
//
//   Free a parameter cell.
//
// Global Variables: -
//
// Side Effects    : Memory Operations
//
/----------------------------------------------------------------------*/

void HeuristicParmsFree(HeuristicParms_p junk)
{
   assert(junk);
   
   junk->heuristic_name = NULL;
   PStackFree(junk->wfcb_definitions);
   PStackFree(junk->heuristic_definitions);
   PStackFree(junk->i_parms);
   DStackFree(junk->d_parms);
   PStackFree(junk->s_parms);

   HeuristicParmsCellFree(junk);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


